/**
  *  ======================================================================
  *  TestDirectedGraph2.java: Create a depencency chart, directed graph,
  *  then locate and print any implicit loops (cycles).
  *  ======================================================================
  */

package demo;

import java.util.Iterator;
import java.util.Set;
 
import org.jgrapht.alg.cycle.CycleDetector;
import org.jgrapht.traverse.TopologicalOrderIterator;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
 
public class DirectedGraphDemo02 {
 
   /**
     *  @param createCycles true - create a directed graph which contains
     * cycles.  false - create a directed graph which does not contain any cycles.
     */

   public static void test(boolean createCycles) {
      CycleDetector<String, DefaultEdge> cycleDetector;
      DefaultDirectedGraph<String, DefaultEdge> g;
 
      g = new DefaultDirectedGraph<String, DefaultEdge>( DefaultEdge.class );
 
      // Add vertices, e.g. equations.

      g.addVertex("a");
      g.addVertex("b");
      g.addVertex("c");
      g.addVertex("d");
      g.addVertex("e");
 
      // Add edges, e.g. dependencies.
      // 2 cycles,
      //    a = f(b)
      //    b = f(c)
      //    c = f(a)
      // and
      //    d = f(e)
      //    e = f(d)

      g.addEdge("b", "a");
      g.addEdge("c", "b");

      if (createCycles) {
          g.addEdge("a", "c");
      }
      g.addEdge("e", "d");
      if (createCycles) {
         g.addEdge("d", "e");
      }

      // Print details of assembled graph
 
      System.out.println( g.toString() );
 
      // Test: Are there cycles in the dependencies?

      cycleDetector = new CycleDetector<String, DefaultEdge>(g);

      // Cycle(s) detected.

      if (cycleDetector.detectCycles()) {
         Iterator<String> iterator;
         Set<String> cycleVertices;
         Set<String> subCycle;
         String cycle;
 
         System.out.println("Cycles detected.");
 
         // Get all vertices involved in cycles.

         cycleVertices = cycleDetector.findCycles();
 
         // Loop through vertices trying to find disjoint cycles.

         while (! cycleVertices.isEmpty()) {
            System.out.println("Cycle:");
 
            // Get a vertex involved in a cycle.
            iterator = cycleVertices.iterator();
            cycle = iterator.next();
 
            // Get all vertices involved with this vertex.
            subCycle = cycleDetector.findCyclesContainingVertex(cycle);
            for (String sub : subCycle) {
               System.out.println("   " + sub);
               // Remove vertex so that this cycle is not encountered
               // again.
               cycleVertices.remove(sub);
            }
         }
      }
 
      // No cycles.  Just output properly ordered vertices.

      else {
         String v;
         TopologicalOrderIterator<String, DefaultEdge> orderIterator;
 
         orderIterator = new TopologicalOrderIterator<String, DefaultEdge>(g);
         System.out.println("\nOrdering:");
         while (orderIterator.hasNext()) {
            v = orderIterator.next();
            System.out.println(v);
         }
      }
   }
 
   public static void main(String [] args) {

      System.out.println("Case 1: Create graph with cycles.");
      System.out.println("================================== ");

      test(true);
 
      System.out.println("Case 2: Create graph with no cycles.");
      System.out.println("=================================== ");

      test(false);
 
      System.out.println("All done");
      System.exit(0);
   }
}
 
