/**
 * =======================================================================
 * MetroSystemGraph.java. Create a directed graph for the Washington DC
 *                        Metro System.
 *
 * Note: A directed graph is called strongly connected if there is a path
 * from each vertex in the graph to every other vertex. In particular,
 * this means paths in each direction; a path from a to b and also a path
 * from b to a.
 *
 * Written by: Mark Austin                                   October, 2011
 * =======================================================================
 */

package demo;

import java.util.List;
import java.util.Set;

import org.jgrapht.*;
import org.jgrapht.alg.*;
import org.jgrapht.alg.connectivity.*;
import org.jgrapht.alg.interfaces.ShortestPathAlgorithm.*;
import org.jgrapht.alg.interfaces.*;
import org.jgrapht.alg.shortestpath.*;
import org.jgrapht.graph.*;

import java.util.*;

import demo.metro.*;
import demo.metro.SymbolTable;

public class MetroSystemGraph {
    String         sName;
    SymbolTable stations;
    DefaultDirectedGraph<MetroStation, DefaultEdge> metro;

    // Constructor methods ....

    public MetroSystemGraph() {
       stations = new SymbolTable();
       metro    = new DefaultDirectedGraph<MetroStation, DefaultEdge> (DefaultEdge.class);
    }

    public MetroSystemGraph( String sName ) {
       this.sName = sName;
       stations = new SymbolTable();
       metro    = new DefaultDirectedGraph<MetroStation, DefaultEdge> (DefaultEdge.class);
    }

    // Create Metro Station objects .....

    public void metroStations() {

        System.out.println("Creating metro stations..." );

        // Create metro stations. (x,y) coordinates measured in miles
        //                        from Union Station.

        MetroStation gA = new MetroStation(           "Greenbelt",   5.0,  12.0 );
        gA.setParking( true );
        MetroStation gB = new MetroStation(        "College Park",   5.0,  10.0 );
        gB.setParking( true );
        MetroStation gC = new MetroStation(       "Silver Spring",  -0.5,  10.0 );
        gC.setParking( true );
        MetroStation gD = new MetroStation(         "Fort Totten",   0.0,   2.0 );
        gD.setParking( true );
        MetroStation gE = new MetroStation(       "Union Station",   0.0,   0.0 );
        gE.setParking( false );
        MetroStation gF = new MetroStation(       "DuPont Circle",  -3.0,   1.0 );
        gF.setParking( false );
        MetroStation gG = new MetroStation( "Catholic University",   0.0,   1.0 );
        gG.setParking( false );
        MetroStation gH = new MetroStation(        "Judiciary Sq",  -1.0,   0.5 );
        gH.setParking( false );
        MetroStation gI = new MetroStation(    "National Airport",  -1.0,  -2.0 );
        gI.setParking( true );

        // Add metro stations along the "orange" line ....

        MetroStation oA = new MetroStation( "New Carrollton",   5.0,   8.0 );
        oA.setParking( true );
        MetroStation oB = new MetroStation( "Stadium-Armory",   1.0,  -1.0 );
        oB.setParking( true );
        MetroStation oC = new MetroStation( "L Enfant Plaza",  -1.0,  -1.0 );
        oC.setParking( false );
        MetroStation oD = new MetroStation(    "Smithsonian",  -1.5,  -0.8 );
        oD.setParking( false );
        MetroStation oE = new MetroStation(   "Metro Center",  -1.5,   0.0 );
        oE.setParking( false );
        MetroStation oF = new MetroStation(        "Rosalyn",  -4.5,   0.0 );
        oF.setParking( false );
        MetroStation oG = new MetroStation(   "Ballston-GMU",  -6.5,   0.0 );
        oG.setParking( false );
        MetroStation oH = new MetroStation(         "Vienna",  -12.0,   0.0 );
        oH.setParking( true );

        // Add metro stations to the metro system database/symbol table.

        stations.put(           "Greenbelt",  gA );
        stations.put(        "College Park",  gB );
        stations.put(       "Silver Spring",  gC );
        stations.put(         "Fort Totten",  gD );
        stations.put(       "Union Station",  gE );
        stations.put(       "DuPont Circle",  gF );
        stations.put( "Catholic University",  gG );
        stations.put(        "Judiciary Sq",  gH );
        stations.put(    "National Airport",  gI );

        // Add stations along the orange line ....

        stations.put(      "New Carrollton",  oA );
        stations.put(      "Stadium-Armory",  oB );
        stations.put(      "L Enfant Plaza",  oC );
        stations.put(         "Smithsonian",  oD );
        stations.put(        "Metro Center",  oE );
        stations.put(             "Rosalyn",  oF );
        stations.put(        "Ballston-GMU",  oG );
        stations.put(              "Vienna",  oH );

        // Define stations along the green and red lines ....

        String redLine[] = {    "Silver Spring",    "Fort Totten",
                          "Catholic University",  "Union Station",
                                 "Metro Center",  "DuPont Circle" };
        String greenLine[] = {      "Greenbelt",   "College Park",
                                  "Fort Totten",   "Judiciary Sq", 
                               "L Enfant Plaza" };
        String yellowLine[] = { "National Airport",
                                  "L Enfant Plaza",
                                    "Judiciary Sq" };
        String orangeLine[] = {   "New Carrollton",  
                                  "Stadium-Armory",  
                                  "L Enfant Plaza", 
                                     "Smithsonian", 
                                    "Metro Center", 
                                         "Rosalyn",  
                                    "Ballston-GMU", 
                                          "Vienna" };

        // Add track assignments to metro station descriptions ....

        for ( int i = 0; i < redLine.length; i = i + 1 ) {
              MetroStation m = (MetroStation) stations.get( redLine[i] );
              m.add("Red");
        }

        for ( int j = 0; j < greenLine.length; j = j + 1 ) {
              MetroStation m = (MetroStation) stations.get( greenLine[j] );
              m.add("Green");
        }

        for ( int j = 0; j < yellowLine.length; j = j + 1 ) {
              MetroStation m = (MetroStation) stations.get( yellowLine[j] );
              m.add("Yellow");
        }

        for ( int j = 0; j < orangeLine.length; j = j + 1 ) {
              MetroStation m = (MetroStation) stations.get( orangeLine[j] );
              m.add("Orange");
        }

        // Transfer symboltable references to directed graph ....

        String names[] = stations.keys();
        for ( int j = 0; j < names.length; j = j + 1 ) {
            MetroStation m = (MetroStation) stations.get( names[j] );
            metro.addVertex( m );
        }
    }

    // Connectivity details for the DC Metrosystem Network

    public void metroNetwork() {
        MetroStation start, end; 

        System.out.println("Creating metro network..." );

        // Create track/links along green line .....

        start = (MetroStation) stations.get(    "Greenbelt" );
        end   = (MetroStation) stations.get( "College Park" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "College Park" );
        end   = (MetroStation) stations.get(  "Fort Totten" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get(  "Fort Totten" );
        end   = (MetroStation) stations.get( "Judiciary Sq" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get(   "Judiciary Sq" );
        end   = (MetroStation) stations.get( "L Enfant Plaza" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        // Create links along red line .....

        start = (MetroStation) stations.get( "Silver Spring");
        end   = (MetroStation) stations.get(  "Fort Totten" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get(  "Fort Totten" );
        end   = (MetroStation) stations.get(  "Catholic University");
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get(  "Catholic University");
        end   = (MetroStation) stations.get(  "Union Station");
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get(  "Union Station");
        end   = (MetroStation) stations.get(  "Metro Center");
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get(    "Metro Center");
        end   = (MetroStation) stations.get(  "DuPont Circle" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        // Create links along yellow line .....

        start = (MetroStation) stations.get(   "Judiciary Sq" );
        end   = (MetroStation) stations.get( "L Enfant Plaza" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "L Enfant Plaza" );
        end   = (MetroStation) stations.get( "National Airport");
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        // Create links along orange line .....

        start = (MetroStation) stations.get( "New Carrollton" );
        end   = (MetroStation) stations.get( "Stadium-Armory" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "Stadium-Armory" );
        end   = (MetroStation) stations.get( "L Enfant Plaza" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "L Enfant Plaza" );
        end   = (MetroStation) stations.get( "Smithsonian" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "Smithsonian" );
        end   = (MetroStation) stations.get( "Metro Center" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "Metro Center" );
        end   = (MetroStation) stations.get( "Rosalyn" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "Rosalyn" );
        end   = (MetroStation) stations.get( "Ballston-GMU" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );

        start = (MetroStation) stations.get( "Ballston-GMU" );
        end   = (MetroStation) stations.get( "Vienna" );
        metro.addEdge( start, end );
        metro.addEdge( end, start );
    }

    // Print details of the Metro System 

    public void print() {
        System.out.println("Washington DC Metro System");
        System.out.println("==========================");
        System.out.println( stations.toString () );
    }   

    // =============================================================
    // Build metro system model and exercise graph algorithms ....
    // =============================================================

    public static void main(String args[]) {

        System.out.println("Create Washington DC Metro System Graph");
        System.out.println("=======================================");

        // Build model of metro stations and rail network ... 

        MetroSystemGraph ms = new MetroSystemGraph("DC Metro");
        ms.metroStations();
        ms.metroNetwork();
        ms.print();

        // Retrieve and print details of individual metro stations ....

        System.out.println("");
        System.out.println("Print Metro Station Details");
        System.out.println("===========================");

        System.out.println( ms.stations.get(      "Greenbelt").toString() );
        System.out.println( ms.stations.get(    "Fort Totten").toString() );
        System.out.println( ms.stations.get( "L Enfant Plaza").toString() );

        // ====================================================
        // Graph Analysis ....
        // ====================================================

        // Part 1. Compute and print all the strongly connected components
        // of the directed graph

        KosarajuStrongConnectivityInspector<MetroStation, DefaultEdge> sci =
              new KosarajuStrongConnectivityInspector<MetroStation, DefaultEdge>( ms.metro );

        List<Set<MetroStation>> stronglyConnectedSubgraphs = sci.stronglyConnectedSets();

        System.out.println("Strongly connected graph components:");
        System.out.println("========================================");
        for (int i = 0; i < stronglyConnectedSubgraphs.size(); i++) {
            System.out.println(stronglyConnectedSubgraphs.get(i));
        }
        System.out.println();

        // Part 2. Prints the shortest path from vertex i to vertex c. This
        // certainly exists for our particular directed graph.

        System.out.println("");
        System.out.println("Shortest path from \"Greenbelt\" to \"National Airport\"");
        System.out.println("========================================================");

        MetroStation start01 = (MetroStation) ms.stations.get( "Greenbelt" );
        MetroStation   end01 = (MetroStation) ms.stations.get( "National Airport");

        GraphPath<MetroStation, DefaultEdge> path01 =
           new DijkstraShortestPath<MetroStation,DefaultEdge>(ms.metro, Double.POSITIVE_INFINITY).getPath(start01, end01);

        System.out.printf("--- Path: %s ...\n", path01.toString() );

        System.out.println("Shortest path from \"Greenbelt\" to \"New Carrollton\"");
        System.out.println("========================================================");

        MetroStation start02 = (MetroStation) ms.stations.get( "Greenbelt" );
        MetroStation   end02 = (MetroStation) ms.stations.get( "New Carrollton");

        GraphPath<MetroStation, DefaultEdge> path02 =
           new DijkstraShortestPath<MetroStation,DefaultEdge>(ms.metro, Double.POSITIVE_INFINITY).getPath(start02, end02);
        System.out.printf("--- Path: %s ...\n", path02.toString() );
    }
}
