/*
 * (C) Copyright 2015-2018, by Graeme Ahokas and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 */

package demo.jgrapht.alg.matching;

import demo.jgrapht.*;

import org.jgrapht.*;
import org.jgrapht.alg.interfaces.MatchingAlgorithm.*;
import org.jgrapht.alg.matching.*;
import org.jgrapht.graph.*;
import org.junit.*;
import org.junit.experimental.categories.*;

import java.math.*;
import java.util.*;

import static org.junit.Assert.*;

public class MaximumWeightBipartiteMatchingTest {

    private SimpleWeightedGraph<String, DefaultWeightedEdge> graph;
    private Set<String> partition1;
    private Set<String> partition2;

    private MaximumWeightBipartiteMatching<String, DefaultWeightedEdge> matcher;

    @Before
    public void setUpGraph() {

        System.out.printf("--- \n");
        System.out.printf("--- Entering MaximumWeightBipartiteMatchingTest.setUpGraph() ...\n");

        graph = new SimpleWeightedGraph<>(DefaultWeightedEdge.class);
        graph.addVertex("s1");
        graph.addVertex("s2");
        graph.addVertex("s3");
        graph.addVertex("s4");
        graph.addVertex("t1");
        graph.addVertex("t2");
        graph.addVertex("t3");
        graph.addVertex("t4");

        partition1 = new HashSet<>();
        partition1.add("s1");
        partition1.add("s2");
        partition1.add("s3");
        partition1.add("s4");

        partition2 = new HashSet<>();
        partition2.add("t1");
        partition2.add("t2");
        partition2.add("t3");
        partition2.add("t4");

        System.out.printf("Graph graph ==> %s ...\n", graph.toString() );
        System.out.printf("Partition 1 ==> %s ...\n", partition1.toString() );
        System.out.printf("Partition 2 ==> %s ...\n", partition2.toString() );

        System.out.printf("--- Leaving MaximumWeightBipartiteMatchingTest.setUpGraph() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void maximumWeightBipartiteMatching1() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching1() ...\n");

        DefaultWeightedEdge e1 = graph.addEdge("s1", "t1");
        graph.setEdgeWeight(e1, 1);
        matcher = new MaximumWeightBipartiteMatching<>(graph, partition1, partition2);
        Matching<String, DefaultWeightedEdge> matchings = matcher.getMatching();

        assertEquals(1, matchings.getEdges().size());
        assertTrue(matchings.getEdges().contains(e1));

        System.out.printf("Matching matchings ==> %s ...\n", matchings.toString() );

        System.out.printf("--- Leave MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching1() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void maximumWeightBipartiteMatching2() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching2() ...\n");

        DefaultWeightedEdge e1 = graph.addEdge("s1", "t1");
        graph.setEdgeWeight(e1, 1);
        DefaultWeightedEdge e2 = graph.addEdge("s2", "t1");
        graph.setEdgeWeight(e2, 2);

        matcher = new MaximumWeightBipartiteMatching<>(graph, partition1, partition2);
        Matching<String, DefaultWeightedEdge> matchings = matcher.getMatching();

        System.out.printf("Graph graph ==> %s ...\n", graph.toString() );
        System.out.printf("Matching matchings ==> %s ...\n", matchings.toString() );

        assertEquals(1, matchings.getEdges().size());
        assertTrue(matchings.getEdges().contains(e2));

        System.out.printf("--- Leave MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching2() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void maximumWeightBipartiteMatching3() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching3() ...\n");

        DefaultWeightedEdge e1 = graph.addEdge("s1", "t1");
        graph.setEdgeWeight(e1, 2);
        DefaultWeightedEdge e2 = graph.addEdge("s1", "t2");
        graph.setEdgeWeight(e2, 1);
        DefaultWeightedEdge e3 = graph.addEdge("s2", "t1");
        graph.setEdgeWeight(e3, 2);

        matcher = new MaximumWeightBipartiteMatching<>(graph, partition1, partition2);
        Matching<String, DefaultWeightedEdge> matchings = matcher.getMatching();

        assertEquals(2, matchings.getEdges().size());
        assertTrue(matchings.getEdges().contains(e2));
        assertTrue(matchings.getEdges().contains(e3));

        System.out.printf("Graph graph ==> %s ...\n", graph.toString() );
        System.out.printf("Matching matchings ==> %s ...\n", matchings.toString() );

        System.out.printf("--- Leave MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching3() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void maximumWeightBipartiteMatching4() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching4() ...\n");

        DefaultWeightedEdge e1 = graph.addEdge("s1", "t1");
        graph.setEdgeWeight(e1, 1);
        DefaultWeightedEdge e2 = graph.addEdge("s1", "t2");
        graph.setEdgeWeight(e2, 1);
        DefaultWeightedEdge e3 = graph.addEdge("s2", "t2");
        graph.setEdgeWeight(e3, 1);

        matcher = new MaximumWeightBipartiteMatching<>(graph, partition1, partition2);
        Matching<String, DefaultWeightedEdge> matchings = matcher.getMatching();
        assertEquals(2, matchings.getEdges().size());
        assertTrue(matchings.getEdges().contains(e1));
        assertTrue(matchings.getEdges().contains(e3));

        System.out.printf("Graph graph ==> %s ...\n", graph.toString() );
        System.out.printf("Matching matchings ==> %s ...\n", matchings.toString() );

        System.out.printf("--- Leave MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching4() ...\n");
        System.out.printf("--- \n");
    }

    @Test
    public void maximumWeightBipartiteMatching5() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching5() ...\n");

        DefaultWeightedEdge e1 = graph.addEdge("s1", "t1");
        graph.setEdgeWeight(e1, 1);
        DefaultWeightedEdge e2 = graph.addEdge("s1", "t2");
        graph.setEdgeWeight(e2, 2);
        DefaultWeightedEdge e3 = graph.addEdge("s2", "t2");
        graph.setEdgeWeight(e3, 2);
        DefaultWeightedEdge e4 = graph.addEdge("s3", "t2");
        graph.setEdgeWeight(e4, 2);
        DefaultWeightedEdge e5 = graph.addEdge("s3", "t3");
        graph.setEdgeWeight(e5, 1);
        DefaultWeightedEdge e6 = graph.addEdge("s4", "t1");
        graph.setEdgeWeight(e6, 1);
        DefaultWeightedEdge e7 = graph.addEdge("s4", "t4");
        graph.setEdgeWeight(e7, 1);

        matcher = new MaximumWeightBipartiteMatching<>(graph, partition1, partition2);
        Matching<String, DefaultWeightedEdge> matchings = matcher.getMatching();
        assertEquals(4, matchings.getEdges().size());
        assertTrue(matchings.getEdges().contains(e1));
        assertTrue(matchings.getEdges().contains(e3));
        assertTrue(matchings.getEdges().contains(e5));
        assertTrue(matchings.getEdges().contains(e7));

        System.out.printf("Graph graph ==> %s ...\n", graph.toString() );
        System.out.printf("Matching matchings ==> %s ...\n", matchings.toString() );

        System.out.printf("--- Leave MaximumWeightBipartiteMatchingTest.maximumWeightBipartiteMatching5() ...\n");
        System.out.printf("--- \n");

    }

    @Test
    @Category(SlowTests.class)
    public void testRandomInstancesFixedSeed() {
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.testRandomInstancesFixedSeed() ...\n");
        testRandomInstance(new Random(17), 10, 0.7, 2);
    }

    @Test
    @Category(SlowTests.class)
    public void testRandomInstances() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.testRandomInstances() ...\n");
        System.out.printf("--- ================================================================== \n");

        Random rng = new Random();
        testRandomInstance(rng,  3, 0.8, 1);
        testRandomInstance(rng,  5, 0.8, 1);

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leaving MaximumWeightBipartiteMatchingTest.testRandomInstances() ...\n");
        System.out.printf("--- \n");
    }

    private void testRandomInstance(Random rng, int n, double p, int repeat) {

        System.out.printf("--- \n");
        System.out.printf("--- Enter MaximumWeightBipartiteMatchingTest.testRandomInstance() ... \n");
        System.out.printf("--- n = %d p = %f repeat = %d ...\n", n, p, repeat );

        for (int a = 0; a < repeat; a++) {

            // generate random bipartite

            System.out.printf("--- Generate random bipartite graph ... \n");

            Graph<Integer, DefaultWeightedEdge> g = new SimpleWeightedGraph<>(DefaultWeightedEdge.class);

            Set<Integer> partitionA = new LinkedHashSet<>(n);
            for (int i = 0; i < n; i++) {
                g.addVertex(i);
                partitionA.add(i);
            }

            Set<Integer> partitionB = new LinkedHashSet<>(n);
            for (int i = 0; i < n; i++) {
                g.addVertex(n + i);
                partitionB.add(n + i);
            }

            System.out.printf("Graph g ==> %s ...\n", g.toString() );
            System.out.printf("Partition A ==> %s ...\n", partitionA.toString() );
            System.out.printf("Partition B ==> %s ...\n", partitionB.toString() );

            // create edges

            System.out.printf("--- Create edges ... \n");

            for (int i = 0; i < n; i++) {
                for (int j = 0; j < n; j++) {
                    // s->t
                    if (rng.nextDouble() < p) {
                        g.addEdge(i, n + j);
                    }
                }
            }

            System.out.printf("Graph g ==> %s ...\n", g.toString() );

            // Assign random weights

            System.out.printf("--- Assign random weights ... \n");

            for (DefaultWeightedEdge e : g.edgeSet()) {
                g.setEdgeWeight(e, 1000 * rng.nextInt());
            }

            System.out.printf("Graph g ==> %s ...\n", g.toString() );

            // Compute maximum weight matching

            System.out.printf("--- Compute maximum weight matching ... \n");

            MaximumWeightBipartiteMatching<Integer, DefaultWeightedEdge> alg =
                   new MaximumWeightBipartiteMatching<>(g, partitionA, partitionB);
            Matching<Integer, DefaultWeightedEdge> matching = alg.getMatching();
            Map<Integer, BigDecimal> pot = alg.getPotentials();
            Comparator<BigDecimal> comparator = Comparator.<BigDecimal> naturalOrder();

            System.out.printf("Maximum Weight matching A ==> %s ...\n", matching.toString() );
            System.out.printf("Potentials pot ==> %s ...\n", pot.toString() );

            // Assert matching

            System.out.printf("--- Assert matching ... \n");

            Map<Integer, Integer> degree = new HashMap<>();
            for (Integer v : g.vertexSet()) {
                degree.put(v, 0);
            }
            for (DefaultWeightedEdge e : matching.getEdges()) {
                Integer s = g.getEdgeSource(e);
                Integer t = g.getEdgeTarget(e);
                degree.put(s, degree.get(s) + 1);
                degree.put(t, degree.get(t) + 1);
            }
            for (Integer v : g.vertexSet()) {
                assertTrue(degree.get(v) <= 1);
            }

            // Assert non-negative potentials

            System.out.printf("--- Assert non-negative potentials ... \n");

            for (Integer v : g.vertexSet()) {
                assertTrue(comparator.compare(pot.get(v), BigDecimal.ZERO) >= 0);
            }

            // Assert non-negative reduced cost for edges

            System.out.printf("--- Assert non-negative reduced cost for edges ... \n");

            for (DefaultWeightedEdge e : g.edgeSet()) {
                Integer s = g.getEdgeSource(e);
                Integer t = g.getEdgeTarget(e);
                BigDecimal w = BigDecimal.valueOf(g.getEdgeWeight(e));
                assertTrue(comparator.compare(w, pot.get(s).add(pot.get(t))) <= 0);
            }

            // Assert tight edges in matching

            System.out.printf("--- Assert tight edges in matching ... \n");

            for (DefaultWeightedEdge e : matching.getEdges()) {
                Integer s = g.getEdgeSource(e);
                Integer t = g.getEdgeTarget(e);
                BigDecimal w = BigDecimal.valueOf(g.getEdgeWeight(e));
                assertTrue(comparator.compare(w, pot.get(s).add(pot.get(t))) == 0);
            }

            // Assert free nodes have zero potential

            System.out.printf("--- Assert free nodes have zero potential ...\n");

            for (Integer v : g.vertexSet()) {
                if (degree.get(v) == 0) {
                    assertEquals(pot.get(v), BigDecimal.ZERO);
                }
            }
        }
    }
}
