/*
 * (C) Copyright 2016-2018, by Dimitrios Michail and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 */

package demo.jgrapht.alg.matching;

import org.jgrapht.*;
import org.jgrapht.alg.interfaces.*;
import org.jgrapht.alg.interfaces.MatchingAlgorithm.*;
import org.jgrapht.alg.matching.*;
import org.jgrapht.graph.*;
import org.junit.*;

import java.util.*;
import java.util.stream.*;

import static org.junit.Assert.*;

/**
 * Unit tests for the PathGrowingWeightedMatching without heuristics algorithm
 * 
 * @author Dimitrios Michail
 */

public class NoHeuristicsPathGrowingWeightedMatchingTest extends BasePathGrowingWeightedMatchingTest {

    @Override
    public MatchingAlgorithm<Integer, DefaultWeightedEdge> getApproximationAlgorithm(
        Graph<Integer, DefaultWeightedEdge> graph) {
        return new PathGrowingWeightedMatching<>(graph, false);
    };

    @Override
    @Test
    public void testGraph1() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter NoHeuristicsPathGrowingWeightedMatchingTest.testGraph1() ... \n");
        System.out.printf("--- =================================================================== ...\n");

        WeightedPseudograph<Integer, DefaultWeightedEdge> g = new WeightedPseudograph<>(DefaultWeightedEdge.class);

        Graphs.addAllVertices(g, IntStream.range(0, 15).boxed().collect(Collectors.toList()));
        Graphs.addEdge(g, 0, 1, 5.0);
        Graphs.addEdge(g, 1, 2, 2.5);
        Graphs.addEdge(g, 2, 3, 5.0);
        Graphs.addEdge(g, 3, 4, 2.5);
        Graphs.addEdge(g, 4, 0, 2.5);
        Graphs.addEdge(g, 0, 13, 2.5);
        Graphs.addEdge(g, 13, 14, 5.0);
        Graphs.addEdge(g, 1, 11, 2.5);
        Graphs.addEdge(g, 11, 12, 5.0);
        Graphs.addEdge(g, 2, 9, 2.5);
        Graphs.addEdge(g, 9, 10, 5.0);
        Graphs.addEdge(g, 3, 7, 2.5);
        Graphs.addEdge(g, 7, 8, 5.0);
        Graphs.addEdge(g, 4, 5, 2.5);
        Graphs.addEdge(g, 5, 6, 5.0);

        System.out.printf("Graph g ==> %s ...\n", g.toString() );

        MatchingAlgorithm<Integer, DefaultWeightedEdge> mm = getApproximationAlgorithm(g);
        Matching<Integer, DefaultWeightedEdge> m = mm.getMatching();

        System.out.printf("Matching m ==> %s ...\n", m.toString() );

        assertEquals(5, m.getEdges().size());
        assertEquals(22.5, m.getWeight(), MatchingAlgorithm.DEFAULT_EPSILON);
        assertTrue(isMatching(g, m));

        System.out.printf("--- =================================================================== ...\n");
        System.out.printf("--- Leaving NoHeuristicsPathGrowingWeightedMatchingTest.testGraph1() ... \n");
        System.out.printf("--- \n");

    }

    @Override
    @Test
    public void testSelfLoops() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter NoHeuristicsPathGrowingWeightedMatchingTest.testSelfLoops() ... \n");
        System.out.printf("--- ================================================================= ... \n");

        WeightedPseudograph<Integer, DefaultWeightedEdge> g = new WeightedPseudograph<>(DefaultWeightedEdge.class);

        Graphs.addAllVertices(g, Arrays.asList(0, 1, 2, 3));
        Graphs.addEdge(g, 0, 1, 1.0);
        Graphs.addEdge(g, 1, 2, 1.0);
        Graphs.addEdge(g, 2, 3, 1.0);
        Graphs.addEdge(g, 3, 0, 1.0);
        Graphs.addAllVertices(g, Arrays.asList(4, 5, 6, 7));
        Graphs.addEdge(g, 4, 5, 1.0);
        Graphs.addEdge(g, 5, 6, 1.0);
        Graphs.addEdge(g, 6, 7, 1.0);
        Graphs.addEdge(g, 7, 4, 1.0);

        // add self loops
        Graphs.addEdge(g, 0, 0, 100.0);
        Graphs.addEdge(g, 1, 1, 200.0);
        Graphs.addEdge(g, 2, 2, -200.0);
        Graphs.addEdge(g, 3, 3, -100.0);
        Graphs.addEdge(g, 4, 4, 0.0);

        System.out.printf("Graph g ==> %s ...\n", g.toString() );

        MatchingAlgorithm<Integer, DefaultWeightedEdge> mm = getApproximationAlgorithm(g);
        Matching<Integer, DefaultWeightedEdge> m = mm.getMatching();

        System.out.printf("Matching m ==> %s ...\n", m.toString() );

        assertEquals(3, m.getEdges().size());
        assertEquals(3.0, m.getWeight(), MatchingAlgorithm.DEFAULT_EPSILON);
        assertTrue(isMatching(g, m));

        System.out.printf("--- =================================================================== ... \n");
        System.out.printf("--- Leaving NoHeuristicsPathGrowingWeightedMatchingTest.testSelfLoops() ... \n");
        System.out.printf("--- \n");
    }

    @Override
    @Test
    public void test3over4Approximation() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter NoHeuristicsPathGrowingWeightedMatchingTest.test3over4Approximation() ... \n");
        System.out.printf("--- =========================================================================== ...\n");

        WeightedPseudograph<Integer, DefaultWeightedEdge> g = new WeightedPseudograph<>(DefaultWeightedEdge.class);

        Graphs.addAllVertices(g, Arrays.asList(0, 1, 2, 3));
        Graphs.addEdge(g, 0, 1, 1.0);
        Graphs.addEdge(g, 1, 2, 1.0);
        Graphs.addEdge(g, 2, 3, 1.0);
        Graphs.addEdge(g, 3, 0, 1.0);
        Graphs.addAllVertices(g, Arrays.asList(4, 5, 6, 7));
        Graphs.addEdge(g, 4, 5, 1.0);
        Graphs.addEdge(g, 5, 6, 1.0);
        Graphs.addEdge(g, 6, 7, 1.0);
        Graphs.addEdge(g, 7, 4, 1.0);

        System.out.printf("Graph g ==> %s ...\n", g.toString() );

        MatchingAlgorithm<Integer, DefaultWeightedEdge> mm = new PathGrowingWeightedMatching<>(g, false);
        Matching<Integer, DefaultWeightedEdge> m = mm.getMatching();

        System.out.printf("Matching m ==> %s ...\n", m.toString() );

        // maximum here is 4.0
        // path growing algorithm gets 3.0

        assertEquals(3, m.getEdges().size());
        assertEquals(3.0, m.getWeight(), MatchingAlgorithm.DEFAULT_EPSILON);
        assertTrue(isMatching(g, m));

        System.out.printf("--- =========================================================================== ...\n");
        System.out.printf("--- Leave NoHeuristicsPathGrowingWeightedMatchingTest.test3over4Approximation() ... \n");
        System.out.printf("--- \n");
    }

    @Override
    @Test
    public void testMultiGraph() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter NoHeuristicsPathGrowingWeightedMatchingTest.testMultiGraph() ... \n");
        System.out.printf("--- =================================================================== ...\n");

        WeightedPseudograph<Integer, DefaultWeightedEdge> g = new WeightedPseudograph<>(DefaultWeightedEdge.class);

        Graphs.addAllVertices(g, Arrays.asList(0, 1, 2, 3));
        Graphs.addEdge(g, 0, 1, 1.0);
        Graphs.addEdge(g, 1, 2, 1.0);
        Graphs.addEdge(g, 2, 3, 1.0);
        Graphs.addEdge(g, 3, 0, 1.0);
        Graphs.addAllVertices(g, Arrays.asList(4, 5, 6, 7));
        Graphs.addEdge(g, 4, 5, 1.0);
        Graphs.addEdge(g, 5, 6, 1.0);
        Graphs.addEdge(g, 6, 7, 1.0);
        Graphs.addEdge(g, 7, 4, 1.0);

        // add multiple edges

        Graphs.addEdge(g, 0, 1, 2.0);
        Graphs.addEdge(g, 1, 2, 2.0);
        Graphs.addEdge(g, 2, 3, 2.0);
        Graphs.addEdge(g, 3, 0, 2.0);
        Graphs.addEdge(g, 4, 5, 2.0);
        Graphs.addEdge(g, 5, 6, 2.0);
        Graphs.addEdge(g, 6, 7, 2.0);
        Graphs.addEdge(g, 7, 4, 2.0);

        System.out.printf("Graph g ==> %s ...\n", g.toString() );

        MatchingAlgorithm<Integer, DefaultWeightedEdge> mm = new PathGrowingWeightedMatching<>(g, false);
        Matching<Integer, DefaultWeightedEdge> m = mm.getMatching();

        System.out.printf("Matching m ==> %s ...\n", m.toString() );

        // maximum here is 8.0
        // path growing algorithm gets 6.0

        assertEquals(3, m.getEdges().size());
        assertEquals(6.0, m.getWeight(), MatchingAlgorithm.DEFAULT_EPSILON);
        assertTrue( isMatching(g, m) );

        System.out.printf("--- ================================================================== ...\n");
        System.out.printf("--- Leave NoHeuristicsPathGrowingWeightedMatchingTest.testMultiGraph() ... \n");
        System.out.printf("--- \n");
    }

    @Override
    @Test
    public void testDirected() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter NoHeuristicsPathGrowingWeightedMatchingTest.testDirected() ... \n");
        System.out.printf("--- ================================================================ ... \n");

        DirectedWeightedPseudograph<Integer, DefaultWeightedEdge> g =
           new DirectedWeightedPseudograph<>(DefaultWeightedEdge.class);

        Graphs.addAllVertices(g, Arrays.asList(0, 1, 2, 3));
        Graphs.addEdge(g, 0, 1, 1.0);
        Graphs.addEdge(g, 1, 2, 1.0);
        Graphs.addEdge(g, 2, 3, 1.0);
        Graphs.addEdge(g, 3, 0, 1.0);
        Graphs.addAllVertices(g, Arrays.asList(4, 5, 6, 7));
        Graphs.addEdge(g, 4, 5, 1.0);
        Graphs.addEdge(g, 5, 6, 1.0);
        Graphs.addEdge(g, 6, 7, 1.0);
        Graphs.addEdge(g, 7, 4, 1.0);

        // add multiple edges

        Graphs.addEdge(g, 0, 1, 2.0);
        Graphs.addEdge(g, 1, 2, 2.0);
        Graphs.addEdge(g, 2, 3, 2.0);
        Graphs.addEdge(g, 3, 0, 2.0);
        Graphs.addEdge(g, 4, 5, 2.0);
        Graphs.addEdge(g, 5, 6, 2.0);
        Graphs.addEdge(g, 6, 7, 2.0);
        Graphs.addEdge(g, 7, 4, 2.0);

        System.out.printf("Graph g ==> %s ...\n", g.toString() );

        MatchingAlgorithm<Integer, DefaultWeightedEdge> mm = new PathGrowingWeightedMatching<>(g, false);
        Matching<Integer, DefaultWeightedEdge> m = mm.getMatching();

        System.out.printf("Matching m ==> %s ...\n", m.toString() );

        // maximum here is 8.0
        // path growing algorithm gets 6.0

        assertEquals(3, m.getEdges().size());
        assertEquals(6.0, m.getWeight(), MatchingAlgorithm.DEFAULT_EPSILON);
        assertTrue(isMatching(g, m));

        System.out.printf("--- ================================================================ ...\n");
        System.out.printf("--- Enter NoHeuristicsPathGrowingWeightedMatchingTest.testDirected() ... \n");
        System.out.printf("--- \n");
    }
}
