/*
 * ======================================================================
 * (C) Copyright 2016-2018, by Dimitrios Michail and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * ======================================================================
 */

package demo.jgrapht.alg.shortestpath;

import org.jgrapht.*;
import org.jgrapht.alg.interfaces.*;
import org.jgrapht.alg.shortestpath.*;
import org.jgrapht.generate.*;
import org.jgrapht.graph.*;
import org.jgrapht.util.*;
import org.junit.*;

import java.util.*;
import java.util.function.*;

import static org.junit.Assert.*;

/**
 * @author Dimitrios Michail
 */
public class AllPairsShortestPathsTest {

    @Test
    public void testRandomFixedSeed() {
        final long seed = 47;

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllPairsShortestPathsTest.testRandomFixedSeed() ...\n");
        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Seed = %d ...\n", seed );

        Random rng = new Random(seed);
        testAllPairsShortestPaths(rng);

        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Leave AllPairsShortestPathsTest.testRandomFixedSeed() ...\n");
    }

    @Test
    public void testRandomFixedSeed8() {
        final long seed = 8;

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllPairsShortestPathsTest.testRandomFixedSeed() ...\n");
        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Seed = %d ...\n", seed );

        Random rng = new Random(seed);
        testAllPairsShortestPaths(rng);

        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Leave AllPairsShortestPathsTest.testRandomFixedSeed() ...\n");
    }

    @Test
    public void testRandomFixedSeed13() {
        final long seed = 13;

        System.out.printf("--- \n");
        System.out.printf("--- Enter AllPairsShortestPathsTest.testRandomFixedSeed() ...\n");
        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Seed = %d ...\n", seed );

        Random rng = new Random(seed);
        testAllPairsShortestPaths(rng);

        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Leave AllPairsShortestPathsTest.testRandomFixedSeed() ...\n");
    }

    private void testAllPairsShortestPaths(Random rng) {
        final int tests = 5;
        final int n = 20;
        final double p = 0.35;
        final int landmarksCount = 2;

        System.out.printf("--- Enter testAllPairsShortestPaths() ...\n");
        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- No tests = %d ...\n", tests );

        System.out.printf("--- List of references to shortest path algorithms ...\n");

        List<Function<Graph<Integer, DefaultWeightedEdge>,
            ShortestPathAlgorithm<Integer, DefaultWeightedEdge>>> algs = new ArrayList<>();
        algs.add((g) -> new DijkstraShortestPath<>(g));
        algs.add((g) -> new BidirectionalDijkstraShortestPath<>(g));
        algs.add((g) -> new AStarShortestPath<>(g, (u, t) -> 0d));
        algs.add((g) -> {
            Integer[] vertices = g.vertexSet().toArray(new Integer[0]);
            Set<Integer> landmarks = new HashSet<>();
            while (landmarks.size() < landmarksCount) {
                landmarks.add(vertices[rng.nextInt(g.vertexSet().size())]);
            }
            return new AStarShortestPath<>(g, new ALTAdmissibleHeuristic<>(g, landmarks));
        });

        System.out.printf("--- Generate random graph ...\n");

        GraphGenerator<Integer, DefaultWeightedEdge, Integer> gen =
            new GnpRandomGraphGenerator<>(n, p, rng, true);

        for (int i = 0; i < tests; i++) {
            System.out.printf("--- Start test: %d ...\n", i );
            System.out.printf("---   Generate graph ...\n" );

            Graph<Integer, DefaultWeightedEdge> g = new DirectedWeightedPseudograph<>(
                SupplierUtil.createIntegerSupplier(), SupplierUtil.DEFAULT_WEIGHTED_EDGE_SUPPLIER);
            gen.generateGraph(g);

            // Assign random weights

            System.out.printf("---   Assign random weights ...\n" );

            for (DefaultWeightedEdge e : g.edgeSet()) {
                g.setEdgeWeight(e, rng.nextDouble());
            }

            double[][] dist = new double[n][n];

            System.out.printf("---   Loop over algorithms ...\n" );

            int j = 0;
            for (Function<Graph<Integer, DefaultWeightedEdge>,
                ShortestPathAlgorithm<Integer, DefaultWeightedEdge>> spProvider : algs) {
                ShortestPathAlgorithm<Integer, DefaultWeightedEdge> alg = spProvider.apply(g);
                for (Integer v : g.vertexSet()) {
                    for (Integer u : g.vertexSet()) {
                        GraphPath<Integer, DefaultWeightedEdge> path = alg.getPath(v, u);

                        double d;
                        if (path == null) {
                            d = Double.POSITIVE_INFINITY;
                        } else {
                            d = path.getWeight();
                        }

                        if (j == 0) {
                            dist[v][u] = d;
                        } else {
                            assertEquals(dist[v][u], d, 1e-9);
                        }
                    }
                }
                j++;
            }
        }

        System.out.printf("--- ===================================================== ...\n");
        System.out.printf("--- Leave testAllPairsShortestPaths() ...\n");
    }

}
