/*
 * =====================================================================
 * (C) Copyright 2003-2018, by Barak Naveh and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * =====================================================================
 */

package demo.jgrapht.graph;

import org.jgrapht.*;
import org.jgrapht.graph.*;
import org.jgrapht.graph.specifics.*;
import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

/**
 * A unit test for directed multigraph.
 *
 * @author Barak Naveh
 * @since Aug 9, 2003
 */

public class DirectedGraphTest {

    // ~ Instance fields --------------------------------------------------------

    private String v1 = "v1";
    private String v2 = "v2";
    private String v3 = "v3";

    // ~ Methods ----------------------------------------------------------------

    /**
     * .
     */

    @Test
    public void testEdgeSetFactory() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.DirectedGraphTest.testEdgeSetFactory() \n");
        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- initialize test graph ... \n");

        DirectedMultigraph<String, DefaultEdge> g = new LinkedHashSetDirectedMultigraph<>(DefaultEdge.class);

        g.addVertex(v1);
        g.addVertex(v2);
        g.addVertex(v3);

        DefaultEdge e1 = g.addEdge(v1, v2);
        DefaultEdge e2 = g.addEdge(v2, v1);
        DefaultEdge e3 = g.addEdge(v2, v3);
        DefaultEdge e4 = g.addEdge(v3, v1);

        System.out.printf("--- assembled graph ... \n");

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        System.out.printf("--- traverse set of edges ... \n");

        Iterator<DefaultEdge> iter = g.edgeSet().iterator();
        assertEquals(e1, iter.next());
        assertEquals(e2, iter.next());
        assertEquals(e3, iter.next());
        assertEquals(e4, iter.next());
        assertFalse(iter.hasNext());

        assertEquals("([v1, v2, v3], [(v1,v2), (v2,v1), (v2,v3), (v3,v1)])", g.toString());

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.DirectedGraphTest.testEdgeSetFactory() \n");
    }

    @Test
    public void testEdgeOrderDeterminism() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.DirectedGraphTest.testEdgeOrderDeterminism() \n");
        System.out.printf("--- ===================================================================== \n");
        System.out.printf("--- initialize test graph ... \n");

        Graph<String, DefaultEdge> g = new DirectedMultigraph<>(DefaultEdge.class);
        g.addVertex(v1);
        g.addVertex(v2);
        g.addVertex(v3);

        DefaultEdge e1 = g.addEdge(v1, v2);
        DefaultEdge e2 = g.addEdge(v2, v3);
        DefaultEdge e3 = g.addEdge(v3, v1);

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        Iterator<DefaultEdge> iter = g.edgeSet().iterator();
        assertEquals(e1, iter.next());
        assertEquals(e2, iter.next());
        assertEquals(e3, iter.next());

        // some bonus tests

        System.out.printf("--- test edge-vertex incidence ... \n");

        assertTrue(Graphs.testIncidence(g, e1, v1));
        assertTrue(Graphs.testIncidence(g, e1, v2));
        assertFalse(Graphs.testIncidence(g, e1, v3));

        System.out.printf("--- test get oposite vertex ... \n");

        assertEquals(v2, Graphs.getOppositeVertex(g, e1, v1));
        assertEquals(v1, Graphs.getOppositeVertex(g, e1, v2));

        assertEquals("([v1, v2, v3], [(v1,v2), (v2,v3), (v3,v1)])", g.toString());

        System.out.printf("--- ===================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.DirectedGraphTest.testEdgeOrderDeterminism() \n");
    }

    /**
     * .
     */
    @Test
    public void testEdgesOf() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.ListenableGraphTest.testEdgesOf() \n");
        System.out.printf("--- ===================================================================== \n");

        Graph<String, DefaultEdge> g = createMultiTriangle();

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        assertEquals(3, g.edgesOf(v1).size());
        assertEquals(3, g.edgesOf(v2).size());
        assertEquals(2, g.edgesOf(v3).size());

        System.out.printf("--- Test: %s --> %d ... \n", "g.edgesOf(v1).size()", g.edgesOf(v1).size());
        System.out.printf("--- Test: %s --> %d ... \n", "g.edgesOf(v2).size()", g.edgesOf(v2).size());
        System.out.printf("--- Test: %s --> %d ... \n", "g.edgesOf(v3).size()", g.edgesOf(v3).size());

        System.out.printf("--- ===================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.ListenableGraphTest.testEdgesOf() \n");
    }

    /**
     * .
     */
    @Test
    public void testInDegreeOf() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.ListenableGraphTest.testInDegreeOf() \n");
        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- initialize test graph ... \n");

        Graph<String, DefaultEdge> g = createMultiTriangle();

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        assertEquals(2, g.inDegreeOf(v1));
        assertEquals(1, g.inDegreeOf(v2));
        assertEquals(1, g.inDegreeOf(v3));

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.ListenableGraphTest.testInDegreeOf() \n");
    }

    /**
     * .
     */
    @Test
    public void testOutDegreeOf() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.ListenableGraphTest.testOutDegreeOf() \n");
        System.out.printf("--- ==================================================================== \n");

        Graph<String, DefaultEdge> g = createMultiTriangle();

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        assertEquals(1, g.outDegreeOf(v1));
        assertEquals(2, g.outDegreeOf(v2));
        assertEquals(1, g.outDegreeOf(v3));

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.ListenableGraphTest.testOutDegreeOf() \n");

    }

    /**
     * .
     */
    @Test
    public void testVertexOrderDeterminism()
    {
        Graph<String, DefaultEdge> g = createMultiTriangle();
        Iterator<String> iter = g.vertexSet().iterator();
        assertEquals(v1, iter.next());
        assertEquals(v2, iter.next());
        assertEquals(v3, iter.next());
    }

    // Create multi-edge triangle ....

    private Graph<String, DefaultEdge> createMultiTriangle() {
        Graph<String, DefaultEdge> g = new DirectedMultigraph<>(DefaultEdge.class);
        initMultiTriangle(g);

        return g;
    }

    private void initMultiTriangle(Graph<String, DefaultEdge> g) {

        g.addVertex(v1);
        g.addVertex(v2);
        g.addVertex(v3);

        g.addEdge(v1, v2);
        g.addEdge(v2, v1);
        g.addEdge(v2, v3);
        g.addEdge(v3, v1);
    }

    // ~ Inner Classes ----------------------------------------------------------

    /**
     * A graph implementation with an edge factory using linked hash sets.
     * 
     * @param <V> the graph vertex type
     * @param <E> the graph edge type
     */

    private class LinkedHashSetDirectedMultigraph<V, E> extends DirectedMultigraph<V, E> {
        private static final long serialVersionUID = -1826738982402033648L;

        public LinkedHashSetDirectedMultigraph(Class<? extends E> edgeClass) {
            super(edgeClass);
        }

        @Override
        protected Specifics<V, E> createSpecifics(boolean directed) {
            return new FastLookupDirectedSpecifics<>(
                this, new LinkedHashMap<>(), v -> new LinkedHashSet<>());
        }
    }
}
