/*
 * ======================================================================
 * (C) Copyright 2015-2018, by Andrew Chen and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * ======================================================================
 */

package demo.jgrapht.graph.builder;

import org.jgrapht.*;
import org.jgrapht.graph.*;
import org.jgrapht.graph.builder.*;
import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

public class GraphBuilderTest {

    // ~ Instance fields --------------------------------------------------------

    private String v1 = "v1";
    private String v2 = "v2";
    private String v3 = "v3";
    private String v4 = "v4";
    private String v5 = "v5";
    private String v6 = "v6";
    private String v7 = "v7";
    private String v8 = "v8";

    @Test
    public void testAddVertex() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.builder.GraphBuilderTest.testAddVertex() ...\n");
        System.out.printf("--- ===================================================================== \n");
        System.out.printf("--- initialize test graph ... \n");

        Graph<String, DefaultEdge> g =
            new GraphBuilder<>(new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class))
                .addVertex(v1).addVertices(v2, v3).build();

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        assertEquals(3, g.vertexSet().size());
        assertEquals(0, g.edgeSet().size());
        assertTrue(g.vertexSet().containsAll(Arrays.asList(v1, v2, v3)));

        System.out.printf("--- ===================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.builder.GraphBuilderTest.testAddVertex() ...\n");
    }

    @Test
    public void testAddEdge() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.builder.GraphBuilderTest.testAddEdge() ...\n");
        System.out.printf("--- =================================================================== \n");

        DefaultWeightedEdge e1 = new DefaultWeightedEdge();
        DefaultWeightedEdge e2 = new DefaultWeightedEdge();

        Graph<String,
            DefaultWeightedEdge> g = new GraphBuilder<>(
                new DefaultDirectedWeightedGraph<String, DefaultWeightedEdge>(
                    DefaultWeightedEdge.class))
                        .addEdge(v1, v2).addEdgeChain(v3, v4, v5, v6).addEdge(v7, v8, 10.0)
                        .addEdge(v1, v7, e1).addEdge(v1, v8, e2, 42.0).buildAsUnmodifiable();

        System.out.printf("--- Graph: %s ... \n", g.toString() );

        assertEquals(8, g.vertexSet().size());
        assertEquals(7, g.edgeSet().size());
        assertTrue(g.vertexSet().containsAll(Arrays.asList(v1, v2, v3, v4, v5, v6, v7, v8)));
        assertTrue(g.containsEdge(v1, v2));
        assertTrue(g.containsEdge(v3, v4));
        assertTrue(g.containsEdge(v4, v5));
        assertTrue(g.containsEdge(v5, v6));
        assertTrue(g.containsEdge(v7, v8));
        assertTrue(g.containsEdge(v1, v7));
        assertTrue(g.containsEdge(v1, v8));
        assertEquals(e1, g.getEdge(v1, v7));
        assertEquals(e2, g.getEdge(v1, v8));
        assertEquals(10.0, g.getEdgeWeight(g.getEdge(v7, v8)), 0);
        assertEquals(42.0, g.getEdgeWeight(g.getEdge(v1, v8)), 0);

        System.out.printf("--- =================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.builder.GraphBuilderTest.testAddEdge() ...\n");
    }

    @Test
    public void testAddGraph() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.builder.GraphBuilderTest.testAddGraph() ...\n");
        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Build graph g1 ... \n");

        Graph<String,
            DefaultEdge> g1 = DefaultDirectedGraph
                .<String, DefaultEdge> createBuilder(DefaultEdge.class).addVertex(v1)
                .addEdge(v2, v3).buildAsUnmodifiable();

        System.out.printf("--- Add graph g1 to graph g2 ... \n");

        System.out.printf("--- Graph g1: %s ... \n", g1.toString() );

        Graph<String, DefaultEdge> g2 =
            new GraphBuilder<>(new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class))
                .addGraph(g1).addEdge(v1, v4).build();

        System.out.printf("--- Graph g2: %s ... \n", g2.toString() );

        assertEquals(4, g2.vertexSet().size());
        assertEquals(2, g2.edgeSet().size());
        assertTrue(g2.vertexSet().containsAll(Arrays.asList(v1, v2, v3, v3)));
        assertTrue(g2.containsEdge(v2, v3));
        assertTrue(g2.containsEdge(v1, v4));

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.builder.GraphBuilderTest.testAddGraph() ...\n");
    }

    @Test
    public void testRemoveVertex() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter demo.jgrapht.graph.builder.GraphBuilderTest.testRemoveVertex() \n");
        System.out.printf("--- ==================================================================== \n");

        Graph<String, DefaultEdge> g1 =
            new GraphBuilder<>(new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class))
                .addEdge(v1, v3).addEdgeChain(v2, v3, v4, v5).buildAsUnmodifiable();

        System.out.printf("--- Graph g1: %s ... \n", g1.toString() );

        Graph<String, DefaultEdge> g2 =
            new GraphBuilder<>(new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class))
                .addGraph(g1).removeVertex(v2).removeVertices(v4, v5).build();

        System.out.printf("--- Graph g2: %s ... \n", g2.toString() );

        assertEquals(2, g2.vertexSet().size());
        assertEquals(1, g2.edgeSet().size());
        assertTrue(g2.vertexSet().containsAll(Arrays.asList(v1, v3)));
        assertTrue(g2.containsEdge(v1, v3));

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave demo.jgrapht.graph.builder.GraphBuilderTest.testRemoveVertex() \n");
    }

    @Test
    public void testRemoveEdge() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter graph.builder.GraphBuilderTest.testRemoveEdge() ... \n");
        System.out.printf("--- ==================================================================== \n");

        DefaultEdge e = new DefaultEdge();

        Graph<String, DefaultEdge> g1 =
            new GraphBuilder<>(new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class))
                .addEdgeChain(v1, v2, v3, v4).addEdge(v1, v4, e).buildAsUnmodifiable();

        System.out.printf("--- Graph g1: %s ... \n", g1.toString() );

        System.out.printf("--- Remove edgeEdge(v2, v3) ... \n");
        System.out.printf("--- Remove edgeEdge(v1, v4, e) ... \n");

        Graph<String, DefaultEdge> g2 =
            new GraphBuilder<>(new DefaultDirectedGraph<String, DefaultEdge>(DefaultEdge.class))
                .addGraph(g1).removeEdge(v2, v3).removeEdge(e).build();

        System.out.printf("--- Graph g2: %s ... \n", g2.toString() );

        assertEquals(4, g2.vertexSet().size());
        assertEquals(2, g2.edgeSet().size());
        assertTrue(g2.vertexSet().containsAll(Arrays.asList(v1, v2, v3, v4)));
        assertTrue(g2.containsEdge(v1, v2));
        assertTrue(g2.containsEdge(v3, v4));

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave graph.builder.GraphBuilderTest.testRemoveEdge() ... \n");
    }

    @Test
    public void testAddVertexPseudograph() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter graph.builder.GraphBuilderTest.testAddVertexPseudograph() ... \n");
        System.out.printf("--- ==================================================================== \n");

        Pseudograph<String, DefaultEdge> g = Pseudograph
            .<String, DefaultEdge> createBuilder(DefaultEdge.class).addVertex(v1).build();

        System.out.printf("--- Graph g: %s ... \n", g.toString() );

        assertEquals(1, g.vertexSet().size());
        assertEquals(0, g.edgeSet().size());
        assertTrue(g.vertexSet().containsAll(Collections.singletonList(v1)));

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Leave graph.builder.GraphBuilderTest.testAddVertexPseudograph() ... \n");
    }

}

// End GraphBuilderTest.java
