/*
 * =========================================================================
 * (C) Copyright 2005-2018, by John V Sichi and Contributors.
 *
 * JGraphT : a free Java graph-theory library
 *
 * This program and the accompanying materials are dual-licensed under
 * either
 *
 * (a) the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation, or (at your option) any
 * later version.
 *
 * or (per the licensee's choosing)
 *
 * (b) the terms of the Eclipse Public License v1.0 as published by
 * the Eclipse Foundation.
 * =========================================================================
 */

package demo.jgrapht.traverse;

import org.jgrapht.*;
import org.jgrapht.graph.*;
import org.jgrapht.traverse.*;
import org.junit.*;

import java.util.*;

import static org.junit.Assert.*;

/**
 * Tests for TopologicalOrderIterator.
 *
 * @author John V. Sichi
 * @since Apr 25, 2005
 */

public class TopologicalOrderIteratorTest {

    @Test
    public void testRecipe() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testRecipe() ... \n");
        System.out.printf("--- ============================================================ \n");

        Graph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);

        System.out.printf("--- Define activities ...\n");

        String[] v = new String[9];

        v[0] = "preheat oven";
        v[1] = "sift dry ingredients";
        v[2] = "stir wet ingredients";
        v[3] = "mix wet and dry ingredients";
        v[4] = "spoon onto pan";
        v[5] = "bake";
        v[6] = "cool";
        v[7] = "frost";
        v[8] = "eat";

        // Add in mixed up order

        graph.addVertex(v[4]);
        graph.addVertex(v[8]);
        graph.addVertex(v[1]);
        graph.addVertex(v[3]);
        graph.addVertex(v[7]);
        graph.addVertex(v[6]);
        graph.addVertex(v[0]);
        graph.addVertex(v[2]);
        graph.addVertex(v[5]);

        System.out.printf("--- Specify dependencies (edges) ... \n");

        // Specify enough edges to guarantee deterministic total order

        graph.addEdge(v[0], v[1]);
        graph.addEdge(v[1], v[2]);
        graph.addEdge(v[0], v[2]);
        graph.addEdge(v[1], v[3]);
        graph.addEdge(v[2], v[3]);
        graph.addEdge(v[3], v[4]);
        graph.addEdge(v[4], v[5]);
        graph.addEdge(v[5], v[6]);
        graph.addEdge(v[6], v[7]);
        graph.addEdge(v[7], v[8]);
        graph.addEdge(v[6], v[8]);

        System.out.printf("--- Graph: %s ... \n", graph.toString() );

        System.out.printf("--- Create TopologicalOrder Iterator ... \n");

        Iterator<String> iter = new TopologicalOrderIterator<>(graph);

        int i = 0;
        System.out.printf("--- Results ...\n");
        System.out.printf("--- ============================ \n");
        while (iter.hasNext()) {
            System.out.printf("--- %s ...\n", iter.next() );
            ++i;
        }
        System.out.printf("--- ============================ \n");

        // Test with a reversed view

        System.out.printf("--- Test with edge reversed view ... \n");

        Graph<String, DefaultEdge> reversed = new EdgeReversedGraph<>(graph);

        iter = new TopologicalOrderIterator<>(reversed);
        i = v.length - 1;

        System.out.printf("--- \n");
        System.out.printf("--- Results ...\n");
        System.out.printf("--- ============================ \n");
        while (iter.hasNext()) {
           System.out.printf("--- %s ...\n", iter.next() );
           --i;
        }

        System.out.printf("--- ============================================================ \n");
        System.out.printf("--- Leave traverse.TopologicalOrderIteratorTest.testRecipe() ... \n");
    }

    @Test
    public void testEmptyGraph() {
        Graph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);
        Iterator<String> iter = new TopologicalOrderIterator<>(graph);
        assertFalse(iter.hasNext());
    }

    @Test
    public void testGraph1() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testGraph1() ... \n");
        System.out.printf("--- ============================================================ \n");

        Graph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);

        Graphs.addAllVertices(graph, Arrays.asList("v0", "v1", "v2", "v3", "v4", "v5"));
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v2");
        graph.addEdge("v1", "v4");
        graph.addEdge("v2", "v4");
        graph.addEdge("v3", "v2");
        graph.addEdge("v3", "v4");
        graph.addEdge("v4", "v5");

        System.out.printf("--- Graph: %s ... \n", graph.toString() );

        Iterator<String> it = new TopologicalOrderIterator<>(graph);

        System.out.printf("--- Results ...\n");
        System.out.printf("--- ============================ \n");
        while (it.hasNext()) {
            System.out.printf("--- %s ...\n", it.next() );
        }

        System.out.printf("--- ============================================================ \n");
        System.out.printf("--- Leave traverse.TopologicalOrderIteratorTest.testGraph1() ... \n");
    }

    @Test
    public void testGraphWithPartialOrder() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testGraphWithPartialOrder() ... \n");
        System.out.printf("--- =========================================================================== \n");

        Graph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);

        Graphs.addAllVertices(graph, Arrays.asList("v0", "v1", "v2", "v3", "v4", "v5"));
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v2");
        graph.addEdge("v1", "v4");
        graph.addEdge("v2", "v4");
        graph.addEdge("v3", "v2");
        graph.addEdge("v3", "v4");
        graph.addEdge("v4", "v5");

        Comparator<String> cf = (a, b) -> {
            if (a.equals("v0") && b.equals("v3")) {
                return 1;
            } else if (a.equals("v3") && b.equals("v0")) {
                return -1;
            }
            if (a.equals("v1") && b.equals("v2")) {
                return 1;
            } else if (a.equals("v2") && b.equals("v1")) {
                return -1;
            }
            return -1;
        };

        Iterator<String> it = new TopologicalOrderIterator<>(graph, cf);

        System.out.printf("--- Results ...\n");
        System.out.printf("--- ============================ \n");
        while (it.hasNext()) {
            System.out.printf("--- %s ...\n", it.next() );
        }

        System.out.printf("--- ============================================================ \n");
        System.out.printf("--- Leave traverse.TopologicalOrderIteratorTest.testGraphWithPartialOrder() ... \n");
    }

    @Test
    public void testGraphWithParallelEdges() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testGraphWithParallelEdges() ... \n");
        System.out.printf("--- ============================================================================ \n");

        Graph<String, DefaultEdge> graph = new DirectedPseudograph<>(DefaultEdge.class);

        Graphs.addAllVertices(graph, Arrays.asList("v0", "v1", "v2", "v3", "v4", "v5"));
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v2");
        graph.addEdge("v1", "v4");
        graph.addEdge("v2", "v4");
        graph.addEdge("v2", "v4");
        graph.addEdge("v3", "v2");
        graph.addEdge("v3", "v4");
        graph.addEdge("v4", "v5");

        System.out.printf("--- Graph: %s ... \n", graph.toString() );

        Iterator<String> it = new TopologicalOrderIterator<>(graph);

        System.out.printf("--- Results ...\n");
        System.out.printf("--- ============================ \n");
        while (it.hasNext()) {
            System.out.printf("--- %s ...\n", it.next() );
        }

        System.out.printf("--- ============================================================================ \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testGraphWithParallelEdges() ... \n");
    }

    @Test(expected = IllegalArgumentException.class)
    public void testWithSelfLoops()
    {
        Graph<String, DefaultEdge> graph = new DefaultDirectedGraph<>(DefaultEdge.class);

        Graphs.addAllVertices(graph, Arrays.asList("v0", "v1", "v2"));
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v2");
        graph.addEdge("v1", "v2");
        graph.addEdge("v2", "v2");

        new TopologicalOrderIterator<>(graph);
    }

    @Test(expected = IllegalArgumentException.class)
    public void testGraphWithCycle() {

        System.out.printf("--- \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testGraphWithCycle() ... \n");
        System.out.printf("--- ==================================================================== \n");

        Graph<String, DefaultEdge> graph = new DirectedPseudograph<>(DefaultEdge.class);

        Graphs.addAllVertices(graph, Arrays.asList("v0", "v1", "v2", "v3", "v4", "v5"));
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v1");
        graph.addEdge("v0", "v2");
        graph.addEdge("v1", "v4");
        graph.addEdge("v2", "v4");
        graph.addEdge("v2", "v4");
        graph.addEdge("v3", "v2");
        graph.addEdge("v3", "v4");
        graph.addEdge("v4", "v5");
        graph.addEdge("v5", "v2");

        Iterator<String> it = new TopologicalOrderIterator<>(graph);

        System.out.printf("--- Results ...\n");
        System.out.printf("--- ============================ \n");
        while (it.hasNext()) {
            System.out.printf("--- %s ...\n", it.next() );
        }

        System.out.printf("--- ==================================================================== \n");
        System.out.printf("--- Enter traverse.TopologicalOrderIteratorTest.testGraphWithCycle() ... \n");
    }

    @Test
    public void testDisconnected() {

        Graph<Integer, DefaultEdge> graph = new DirectedPseudograph<>(DefaultEdge.class);

        Graphs.addAllVertices(graph, Arrays.asList(0, 1, 2, 3));
        graph.addEdge(0, 1);
        graph.addEdge(2, 3);

        Comparator<Integer> cf = (a, b) -> {
            if (a < b) {
                return 1;
            } else if (a > b) {
                return -1;
            } else {
                return 0;
            }
        };

        Iterator<Integer> it = new TopologicalOrderIterator<>(graph, cf);
        assertTrue(it.hasNext());
        assertEquals(Integer.valueOf(2), it.next());
        assertTrue(it.hasNext());
        assertEquals(Integer.valueOf(3), it.next());
        assertTrue(it.hasNext());
        assertEquals(Integer.valueOf(0), it.next());
        assertTrue(it.hasNext());
        assertEquals(Integer.valueOf(1), it.next());
        assertFalse(it.hasNext());
    }

    @Test(expected = IllegalArgumentException.class)
    public void testTryToDisableCrossComponent() {
        Graph<Integer, DefaultEdge> graph = new DirectedPseudograph<>(DefaultEdge.class);
        new TopologicalOrderIterator<>(graph).setCrossComponentTraversal(false);
    }

}
